"""
Simple Log.

Licensed under MIT
Copyright (c) 2013 - 2015 Isaac Muse <isaacmuse@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions
of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
IN THE SOFTWARE.
"""
from __future__ import unicode_literals
import codecs
import threading
from .. import util

ALL = 0
DEBUG = 10
INFO = 20
WARNING = 30
ERROR = 40
CRITICAL = 50

global_log = None


class Log(object):
    """Log."""

    def __init__(self, filename=None, fmt="%(message)s", level=ERROR, filemode="w"):
        """Init Log object."""

        self._lock = threading.Lock()
        if filemode == "w":
            with codecs.open(filename, "w", "utf-8") as f:
                f.write("")
        self.filename = filename
        self.level = level
        self.format = fmt
        self.save_to_file = self.filename is not None
        self.echo = not self.save_to_file

    def set_echo(self, enable):
        """Turn on/off echoing to std out."""

        with self._lock:
            if self.save_to_file:
                self.echo = bool(enable)

    def set_level(self, level):
        """Set log level."""

        with self._lock:
            self.level = int(level)

    def get_level(self):
        """Get log level."""

        return self.level

    def formatter(self, lvl, log_fmt, msg, msg_fmt=None):
        """Special formatters for log message."""

        return log_fmt % {
            "loglevel": lvl,
            "message": util.to_ustr(msg if msg_fmt is None else msg_fmt(msg))
        }

    def debug(self, msg, log_fmt="%(loglevel)s: %(message)s\n", echo=True, msg_fmt=None):
        """Debug level logging."""

        if self.level <= DEBUG:
            self._log(self.formatter("DEBUG", log_fmt, msg, msg_fmt), echo)

    def info(self, msg, log_fmt="%(loglevel)s: %(message)s\n", echo=True, msg_fmt=None):
        """Info level logging."""

        if self.level <= INFO:
            self._log(self.formatter("INFO", log_fmt, msg, msg_fmt), echo)

    def warning(self, msg, log_fmt="%(loglevel)s: %(message)s\n", echo=True, msg_fmt=None):
        """Warning level logging."""

        if self.level <= WARNING:
            self._log(self.formatter("WARNING", log_fmt, msg, msg_fmt), echo)

    def error(self, msg, log_fmt="%(loglevel)s: %(message)s\n", echo=True, msg_fmt=None):
        """Error level logging."""

        if self.level <= ERROR:
            self._log(self.formatter("ERROR", log_fmt, msg, msg_fmt), echo)

    def critical(self, msg, log_fmt="%(loglevel)s: %(message)s\n", echo=True, msg_fmt=None):
        """Critical level logging."""

        if self.level <= CRITICAL:
            self._log(self.formater("CRITICAL", log_fmt, msg, msg_fmt), echo)

    def _log(self, msg, echo=True):
        """Base logger."""

        if not (echo and self.echo) and self.save_to_file:
            with self._lock:
                with codecs.open(self.filename, "a", "utf-8") as f:
                    f.write(self.format % {"message": msg})
        if echo and self.echo:
            print(self.format % {"message": msg})

    def read(self):
        """Read the log."""

        txt = ""
        with self._lock:
            try:
                with codecs.open(self.filename, "r", "utf-8") as f:
                    txt = f.read()
            except Exception:
                pass
        return txt


def init_global_log(file_name, level=ERROR):
    """Initialize a global log object."""

    global global_log
    global_log = Log(file_name, level=level)


def get_global_log():
    """Get the global log object."""

    return global_log



"""
Sublime Text Color Scheme Editor.

Licensed under MIT
Copyright (c) 2013 Isaac Muse <isaacmuse@gmail.com>
"""
from __future__ import unicode_literals
import sys
import codecs
import os
import plistlib
import threading
import time
import uuid
import wx
import traceback

from .about_dialog import AboutDialog
from .custom_app import DebugFrameExtender
from .custom_app import debug, debug_struct, error
from . import basic_dialogs
from . import gui
from . import global_settings_panel
from . import style_settings_panel
from . import data
from . import settings_codes as sc
from ..default_new_theme import theme as default_new_theme
from .. import util

DEBUG_CONSOLE = False

SHORTCUTS = {
    "osx": '''
===Applicatioon Shortcuts===
Find Next: \u2318 + F
Find Next: \u2318 + G
Find Prev: \u2318 + \u21e7 + G
Save: \u2318 + S
Save As: \u2318 + \u21e7 + S

===Table Shortcuts===
Edit Row: Enter
Move Row Up (Style Settings): \u2325 + \u2191
Move Row Down (Style Settings): \u2325 + \u2193
Switch to Global Settings: \u2325 + \u2190
Switch to Style Settings: \u2325 + \u2192
Delete Row: \u232B
Insert Row: \u2318 + I
Toggle 'bold' font: B
Toggle 'italic' font: I
Toggle 'underlined' font: U
''',

    "windows": '''
===Applicatioon Shortcuts===
Find Next: Control + F
Find Next: Control + G
Find Prev: Control + Shift + G
Save: Control + S
Save As: Control + Shift + S

===Table Shortcuts===
Edit Row: Enter
Move Row Up (Style Settings): Alt + \u2191
Move Row Down (Style Settings): Alt + \u2193
Switch to Global Settings: Alt + \u2190
Switch to Style Settings: Alt + \u2192
Delete Row: Delete
Insert Row: Control + I
Toggle 'bold' font: B
Toggle 'italic' font: I
Toggle 'underlined' font: U
''',

    "linux": '''
===Applicatioon Shortcuts===
Find Next: Control + F
Find Next: Control + G
Find Prev: Control + Shift + G
Save: Control + S
Save As: Control + Shift + S

===Table Shortcuts===
Edit Row: Enter
Move Row Up (Style Settings): Alt + \u2191
Move Row Down (Style Settings): Alt + \u2193
Switch to Global Settings: Alt + \u2190
Switch to Style Settings: Alt + \u2192
Delete Row: Delete
Insert Row: Control + I
Toggle 'bold' font: B
Toggle 'italic' font: I
Toggle 'underlined' font: U
'''
}


#################################################
# Helper Functions
#################################################
def query_user_for_file(parent, action):
    """Query the user for file to use."""

    file_path = None
    select_file = action == "select"
    new_file = action == "new"
    select = False
    done = False
    wildcard = "*.tmTheme"
    if not select_file and not new_file:
        select = basic_dialogs.yesno(
            "Create a new theme or select an existing one?", "Color Scheme Editor", yes="Select", no="New"
        )
    elif select_file:
        select = True
    while not done:
        if select:
            result = basic_dialogs.filepicker("Choose a theme file:", "", wildcard)
            if result is not None:
                debug(result)
                if not result.lower().endswith(".tmtheme"):
                    basic_dialogs.errormsg("File must be of type '.tmtheme'")
                    debug("Select: Bad extension: %s" % result)
                    continue
                file_path = result
                debug("Select: File selected: %s" % file_path)
            done = True
        else:
            result = basic_dialogs.filepicker("Theme file to save:", "", wildcard, True)
            if result is not None:
                if not result.lower().endswith(".tmtheme"):
                    basic_dialogs.errormsg("File must be of type '.tmtheme'")
                    debug("New: Bad extension: %s" % result)
                    continue
                try:
                    with codecs.open(result, "w", "utf-8") as f:
                        f.write((util.dump_plist(default_new_theme) + '\n'))
                    file_path = result
                    debug("New: File selected: %s" % file_path)
                except Exception:
                    error(traceback.format_exc())
                    basic_dialogs.errormsg('There was a problem writing the theme file!')
            done = True
    return file_path


def parse_file(file_path):
    """Parse the scheme file."""

    t_file = None
    color_scheme = None

    try:
        with open(file_path, "rb") as f:
            color_scheme = plistlib.readPlist(f)
    except Exception:
        error(traceback.format_exc())
        basic_dialogs.errormsg('Unexpected problem trying to parse file!')

    if color_scheme is not None:
        t_file = file_path

    return t_file, color_scheme


#################################################
# Live Update Manager
#################################################
class LiveUpdate(threading.Thread):
    """
    Live update of the file.

    This is mainly so Sublime can refersh the theme as you edit.
    """

    def __init__(self, func, queue):
        """Initialize."""

        self.func = func
        self.queue = queue
        self.last_queue_len = len(queue)
        self.abort = False
        self.last_update = 0.0
        self.done = False
        self.locked = False
        threading.Thread.__init__(self)

    def kill_thread(self):
        """Kill thread."""

        self.abort = True

    def lock_queue(self):
        """Lock the queue."""

        if not self.is_queue_locked():
            self.locked = True
            return True
        return False

    def release_queue(self):
        """Release the queue."""

        if self.is_queue_locked():
            self.locked = False
            return True
        return False

    def is_queue_locked(self):
        """Check if queue is locked."""

        return self.locked

    def is_done(self):
        """Check if done."""

        return self.done

    def update(self, queue):
        """Update the theme."""

        wx.CallAfter(self.func, "Live Thread")

    def _process_queue(self):
        """Process the queue."""

        while not self.lock_queue():
            time.sleep(.2)
        current_queue = self.queue[0:self.last_queue_len]
        del self.queue[0:self.last_queue_len]
        self.last_queue_len = len(self.queue)
        self.release_queue()
        return current_queue

    def run(self):
        """Run the thread."""

        while not self.abort:
            now = time.time()
            if len(self.queue) and (now - .5) > self.last_update:
                if len(self.queue) != self.last_queue_len:
                    self.last_queue_len = len(self.queue)
                else:
                    self.update(self._process_queue())
                    self.last_update = time.time()
            if self.abort:
                break
            time.sleep(.5)
        if len(self.queue):
            self.update(self._process_queue())
        self.done = True


#################################################
# Editor Dialog
#################################################
class Editor(gui.EditorFrame, DebugFrameExtender):
    """Main editor."""

    def __init__(self, parent, live_save, debugging=False):
        """Initialize."""

        super(Editor, self).__init__(parent)
        self.ready = False
        self.m_global_settings = None
        self.m_style_settings = None
        if util.platform() == "windows":
            self.SetDoubleBuffered(True)
        self.SetIcon(data.get_image('subclrschm_large.png').GetIcon())
        self.live_save = bool(live_save)
        self.updates_made = False
        mod = wx.ACCEL_CMD if sys.platform == "darwin" else wx.ACCEL_CTRL
        self.set_keybindings(
            [
                (mod, ord('B'), self.on_shortcuts),
                (mod | wx.ACCEL_SHIFT, ord('S'), self.on_save_as),
                (mod, ord('S'), self.on_save),
                (mod, ord('F'), self.focus_find),
                (mod, ord('G'), self.on_next_find),
                (mod | wx.ACCEL_SHIFT, ord('G'), self.on_prev_find)
            ],
            self.on_debug_console if debugging else None
        )
        self.search_results = []
        self.cur_search = None
        self.last_UUID = None
        self.last_plist_name = None
        self.m_menuitem_save.Enable(False)
        self.queue = []
        self.debugging = debugging
        self.m_main_panel.Fit()
        self.Fit()
        self.SetMinSize(self.GetSize())

    def init_frame(self, scheme_file, action):
        """Show the main frame."""

        t_file = None
        scheme = None

        if scheme_file is None:
            scheme_file = query_user_for_file(self, action)

        if scheme_file is not None:
            t_file, scheme = parse_file(scheme_file)

        if t_file is None:
            self.Close()
            return

        if self.debugging:
            self.open_debug_console()
        self.SetTitle("Color Scheme Editor - %s" % os.path.basename(t_file))
        self.scheme = scheme
        self.tmtheme = t_file
        debug_struct(scheme, "Color Scheme")

        try:
            self.m_style_settings = style_settings_panel.StyleSettings(
                self.m_plist_notebook, scheme,
                self.update_plist
            )
            self.m_global_settings = global_settings_panel.GlobalSettings(
                self.m_plist_notebook, scheme,
                self.update_plist, self.rebuild_tables
            )
        except Exception as e:
            debug("Failed to load scheme settings!")
            debug(e)
            raise

        self.m_plist_name_textbox.SetValue(scheme["name"])
        self.m_plist_uuid_textbox.SetValue(scheme["uuid"])
        self.last_UUID = scheme["uuid"]
        self.last_plist_name = scheme["name"]

        self.m_plist_notebook.InsertPage(0, self.m_global_settings, "Global Settings", True)
        self.m_plist_notebook.InsertPage(1, self.m_style_settings, "Scope Settings", False)
        if self.live_save:
            self.update_thread = LiveUpdate(self.save, self.queue)
            self.update_thread.start()

        self.ready = True

    def is_ready(self):
        """Get whether dialog is ready to show."""

        return self.ready

    def update_plist(self, code, args={}):
        """Update plist."""

        if code == sc.UUID:
            self.scheme["uuid"] = self.m_plist_uuid_textbox.GetValue()
            self.updates_made = True
        elif code == sc.NAME:
            self.scheme["name"] = self.m_plist_name_textbox.GetValue()
            self.updates_made = True
        elif code == sc.ADD and args is not None:
            debug("Add")
            if args["table"] == "style":
                self.scheme["settings"].insert(args["index"] + 1, args["data"])
            else:
                self.scheme["settings"][0]["settings"][args["index"]] = args["data"]
            self.updates_made = True
        elif code == sc.DELETE and args is not None:
            debug("Delete")
            if args["table"] == "style":
                del self.scheme["settings"][args["index"] + 1]
            else:
                del self.scheme["settings"][0]["settings"][args["index"]]
            self.updates_made = True
        elif code == sc.MOVE and args is not None:
            debug("Move")
            from_row = args["from"] + 1
            to_row = args["to"] + 1
            item = self.scheme["settings"][from_row]
            del self.scheme["settings"][from_row]
            self.scheme["settings"].insert(to_row, item)
            self.updates_made = True
        elif code == sc.MODIFY and args is not None:
            debug("Modify")
            if args["table"] == "style":
                obj = {
                    "name": args["data"]["name"],
                    "scope": args["data"]["scope"],
                    "settings": {
                    }
                }

                settings = args["data"]["settings"]

                if settings["foreground"] != "":
                    obj["settings"]["foreground"] = settings["foreground"]

                if settings["background"] != "":
                    obj["settings"]["background"] = settings["background"]

                if settings["fontStyle"] != "":
                    obj["settings"]["fontStyle"] = settings["fontStyle"]

                self.scheme["settings"][args["index"] + 1] = obj
            else:
                self.scheme["settings"][0]["settings"][args["index"]] = args["data"]
            self.updates_made = True
        else:
            debug("No valid edit actions!")

        if self.live_save:
            while not self.update_thread.lock_queue():
                time.sleep(.2)
            self.queue.append("tmtheme")
            self.update_thread.release_queue()
        elif self.updates_made:
            self.m_menuitem_save.Enable(True)

    def rebuild_plist(self):
        """Rebuild plist."""

        self.scheme["name"] = self.m_plist_name_textbox.GetValue()
        self.scheme["uuid"] = self.m_plist_uuid_textbox.GetValue()
        self.scheme["settings"] = [{"settings": {}}]
        for r in range(0, self.m_global_settings.m_plist_grid.GetNumberRows()):
            key = self.m_global_settings.m_plist_grid.GetCellValue(r, 0)
            val = self.m_global_settings.m_plist_grid.GetCellValue(r, 1)
            self.scheme["settings"][0]["settings"][key] = val

        for r in range(0, self.m_style_settings.m_plist_grid.GetNumberRows()):
            name = self.m_style_settings.m_plist_grid.GetCellValue(r, 0)
            foreground = self.m_style_settings.m_plist_grid.GetCellValue(r, 1)
            background = self.m_style_settings.m_plist_grid.GetCellValue(r, 2)
            fontstyle = self.m_style_settings.m_plist_grid.GetCellValue(r, 3)
            scope = self.m_style_settings.m_plist_grid.GetCellValue(r, 4)

            obj = {
                "name": name,
                "scope": scope,
                "settings": {
                }
            }

            if foreground != "":
                obj["settings"]["foreground"] = foreground

            if background != "":
                obj["settings"]["background"] = background

            if fontstyle != "":
                obj["settings"]["fontStyle"] = fontstyle

            self.scheme["settings"].append(obj)

        if self.live_save:
            while not self.update_thread.lock_queue():
                time.sleep(.2)
            self.queue.append("tmtheme")
            self.update_thread.release_queue()

    def save(self, requester="Main Thread"):
        """Save."""

        debug("%s requested save" % requester)
        try:
            with codecs.open(self.tmtheme, "w", "utf-8") as f:
                f.write((util.dump_plist(self.scheme) + '\n'))
            self.updates_made = False
            if not self.live_save:
                self.m_menuitem_save.Enable(False)
        except Exception:
            error(traceback.format_exc())
            basic_dialogs.errormsg('Unexpected problem trying to write .tmTheme file!')

    def rebuild_tables(self, cur_row, cur_col):
        """Rebuild the tables."""

        cur_page = self.m_plist_notebook.GetSelection()

        self.m_global_settings.m_plist_grid.DeleteRows(0, self.m_global_settings.m_plist_grid.GetNumberRows())
        self.m_global_settings.read_plist(self.scheme)
        self.m_global_settings.go_cell(self.m_global_settings.m_plist_grid, 0, 0)

        self.m_style_settings.m_plist_grid.DeleteRows(0, self.m_style_settings.m_plist_grid.GetNumberRows())
        self.m_style_settings.read_plist(self.scheme)
        self.m_style_settings.go_cell(self.m_style_settings.m_plist_grid, 0, 0)

        if cur_page == 0:
            self.m_plist_notebook.ChangeSelection(cur_page)
            if cur_row is not None and cur_col is not None:
                self.m_global_settings.go_cell(self.m_global_settings.m_plist_grid, cur_row, cur_col, True)
        elif cur_page == 1:
            self.m_plist_notebook.ChangeSelection(cur_page)
            if cur_row is not None and cur_col is not None:
                self.m_style_settings.go_cell(self.m_style_settings.m_plist_grid, cur_row, cur_col, True)

    def set_style_object(self, obj):
        """Set style object."""

        self.m_style_settings.set_object(obj)

    def set_global_object(self, key, value):
        """Set global object."""

        self.m_global_settings.set_object(key, value)

    def focus_find(self, event):
        """Set focus on search panel."""

        self.m_search_panel.SetFocus()
        event.Skip()

    def find(self):
        """Find."""

        self.search_results = []
        pattern = self.m_search_panel.GetValue().lower()
        panel = self.m_style_settings if self.m_plist_notebook.GetSelection() else self.m_global_settings
        self.cur_search = panel
        grid = panel.m_plist_grid
        for r in range(0, grid.GetNumberRows()):
            for c in range(0, grid.GetNumberCols()):
                if pattern in grid.GetCellValue(r, c).lower():
                    self.search_results.append((r, c))

    def find_next(self, current=False):
        """Find next."""

        panel = self.m_style_settings if self.m_plist_notebook.GetSelection() else self.m_global_settings
        if self.cur_search is not panel:
            debug("Find: Panel switched.  Upate results.")
            self.find()
        grid = panel.m_plist_grid
        row = grid.GetGridCursorRow()
        col = grid.GetGridCursorCol()
        next = None
        for i in self.search_results:
            if current and row == i[0] and col == i[1]:
                next = i
                break
            elif row == i[0] and col < i[1]:
                next = i
                break
            elif row < i[0]:
                next = i
                break
        if next is None and len(self.search_results):
            next = self.search_results[0]
        if next is not None:
            grid.SetFocus()
            panel.go_cell(grid, next[0], next[1], True)

    def find_prev(self, current=False):
        """Find previous."""

        panel = self.m_style_settings if self.m_plist_notebook.GetSelection() else self.m_global_settings
        if self.cur_search is not panel:
            debug("Find: Panel switched.  Upate results.")
            self.find()
        grid = panel.m_plist_grid
        row = grid.GetGridCursorRow()
        col = grid.GetGridCursorCol()
        prev = None
        for i in reversed(self.search_results):
            if current and row == i[0] and col == i[1]:
                prev = i
                break
            elif row == i[0] and col > i[1]:
                prev = i
                break
            elif row > i[0]:
                prev = i
                break
        if prev is None and len(self.search_results):
            prev = self.search_results[-1]
        if prev is not None:
            grid.SetFocus()
            panel.go_cell(grid, prev[0], prev[1], True)

    def file_close_cleanup(self):
        """File close cleanup."""

        if self.live_save:
            self.update_thread.kill_thread()
            if self.live_save:
                while not self.update_thread.is_done():
                    time.sleep(0.5)
        elif not self.live_save and self.updates_made:
            if basic_dialogs.yesno("You have unsaved changes.  Save?", "Color Scheme Editor"):
                self.save()

    def on_plist_name_blur(self, event):
        """Handle plist name blur event."""

        event.Skip()
        if not self.is_ready():
            return
        set_name = self.m_plist_name_textbox.GetValue()
        if set_name != self.last_plist_name:
            self.last_plist_name = set_name
            self.update_plist(sc.NAME)

    def on_uuid_button_click(self, event):
        """Handle UUID button event."""

        self.last_UUID = str(uuid.uuid4()).upper()
        self.m_plist_uuid_textbox.SetValue(self.last_UUID)
        self.update_plist(sc.UUID)
        event.Skip()

    def on_uuid_blur(self, event):
        """Handle UUID blur event."""

        event.Skip()
        if not self.is_ready():
            return
        try:
            set_uuid = self.m_plist_uuid_textbox.GetValue()
            uuid.UUID(set_uuid)
            if set_uuid != self.last_UUID:
                self.last_UUID = set_uuid
                self.update_plist(sc.UUID)
        except Exception:
            self.on_uuid_button_click(event)
            error(traceback.format_exc())
            basic_dialogs.errormsg('UUID is invalid! A new UUID has been generated.')
            debug("Bad UUID: %s!" % self.m_plist_uuid_textbox.GetValue())

    def on_plist_notebook_size(self, event):
        """Handle plist notebook size event."""

        if self.m_global_settings and self.m_style_settings:
            self.m_global_settings.resize_table()
            self.m_style_settings.resize_table()
        event.Skip()

    def on_open_new(self, event):
        """Handle open new event."""

        self.file_close_cleanup()
        save_file = query_user_for_file(self, action="select")
        if save_file is not None:
            t_file, color_scheme = parse_file(save_file)
            if t_file is not None:
                self.tmtheme = t_file
                self.SetTitle("Color Scheme Editor - %s" % os.path.basename(t_file))
                self.scheme = color_scheme
                self.m_plist_name_textbox.SetValue(self.scheme["name"])
                self.m_plist_uuid_textbox.SetValue(self.scheme["uuid"])
                self.last_UUID = self.scheme["uuid"]
                self.last_plist_name = self.scheme["name"]
                self.rebuild_tables(None, None)

    def on_save(self, event):
        """Handle save event."""

        if not self.live_save:
            self.save()

    def on_save_as(self, event):
        """Handle save as event."""

        save_file = query_user_for_file(self, action="new")
        if save_file is not None:
            t_file = None
            t_file = save_file
            self.tmtheme = t_file
            self.SetTitle("Color Scheme Editor - %s" % os.path.basename(t_file))
            if self.live_save:
                while not self.update_thread.lock_queue():
                    time.sleep(.2)
                del self.queue[0:len(self.queue)]
                self.update_thread.release_queue()
            self.save()

    def on_about(self, event):
        """Handle about event."""

        dlg = AboutDialog(self)
        dlg.ShowModal()
        dlg.Destroy()

    def on_find(self, event):
        """Handle find event."""

        self.find()
        event.Skip()

    def on_find_finish(self, event):
        """Handle find finish event."""

        self.find_next(current=True)

    def on_next_find(self, event):
        """Handle next find event."""

        self.find_next()

    def on_prev_find(self, event):
        """Handle previous find event."""

        self.find_prev()

    def on_shortcuts(self, event):
        """Handle shortcuts event."""

        if sys.platform == "darwin":
            msg = SHORTCUTS["osx"]
        elif sys.platform == "linux2":
            msg = SHORTCUTS["linux"]
        else:
            msg = SHORTCUTS["windows"]
        basic_dialogs.infomsg(msg, "Shortcuts")

    def on_debug_console(self, event):
        """Handle debug console event."""

        self.open_debug_console()

    def on_close(self, event):
        """Handle close event."""

        self.close_debug_console()
        self.file_close_cleanup()
        event.Skip()



"""Version."""
from __future__ import unicode_literals
from .__version__ import version, version_info

__version__ = version
__version_info__ = version_info
__app__ = "SubClrSchm"
__status__ = version_info[3]
__maintainers__ = [("Isaac Muse", "IsaacMuse@gmail.com")]
__help__ = "https://github.com/facelessuser/subclrschm/issues"
__manual__ = "https://github.com/facelessuser/subclrschm/blob/master/README.md"



"""
Custom App.

https://gist.github.com/facelessuser/5750404

Licensed under MIT
Copyright (c) 2013 - 2015 Isaac Muse <isaacmuse@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions
of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""
from __future__ import unicode_literals
import codecs
import json
import os
import time
import wx
import wx.lib.newevent
from . import simplelog
from .. import util
try:
    import thread
except ImportError:
    import _thread as thread


if util.platform() == "windows":
    import ctypes
    GENERIC_READ = 0x80000000
    GENERIC_WRITE = 0x40000000
    OPEN_EXISTING = 0x3
    PIPE_ACCESS_DUPLEX = 0x3
    PIPE_TYPE_MESSAGE = 0x4
    PIPE_WAIT = 0x0

PipeEvent, EVT_PIPE_ARGS = wx.lib.newevent.NewEvent()

log = None
last_level = simplelog.ERROR
DEBUG_MODE = False
DEBUG_CONSOLE = False


class GuiLog(wx.PyOnDemandOutputWindow):
    """GUI logging."""

    def __init__(self, title="Debug Console"):
        """Init the PyOnDemandOutputWindow object."""

        # wx.PyOnDemandOutputWindow is old class style
        # Cannot use super with old class styles
        wx.PyOnDemandOutputWindow.__init__(self, title)

    def CreateOutputWindow(self, st):  # noqa
        """Create the logging console."""

        self.frame = wx.Frame(self.parent, -1, self.title, self.pos, self.size, style=wx.DEFAULT_FRAME_STYLE)
        self.text = wx.TextCtrl(self.frame, -1, "", style=wx.TE_MULTILINE | wx.TE_READONLY | wx.TE_RICH)
        self.text.AppendText(st)
        self.frame.Show(True)
        self.frame.Bind(wx.EVT_CLOSE, self.OnCloseWindow)

        # Create debug keybinding to open debug console
        debugid = wx.NewId()
        self.frame.Bind(wx.EVT_MENU, self.debug_close, id=debugid)
        mod = wx.ACCEL_CMD if util.platform() == "osx" else wx.ACCEL_CTRL
        accel_tbl = wx.AcceleratorTable(
            [(mod, ord('`'), debugid)]
        )
        self.frame.SetAcceleratorTable(accel_tbl)

    def debug_close(self, event):
        """Close debug frame."""

        self.frame.Close()

    def write(self, text, echo=True):
        """Write to log, and if console is open, echo to it as well."""

        if self.frame is None:
            if not wx.IsMainThread():
                if echo:
                    wx.CallAfter(gui_log, text)
                if get_debug_console():
                    wx.CallAfter(self.CreateOutputWindow, text)
            else:
                if echo:
                    gui_log(text)
                if get_debug_console():
                    self.CreateOutputWindow(text)
        else:
            if not wx.IsMainThread():
                if echo:
                    wx.CallAfter(gui_log, text)
                if get_debug_console():
                    wx.CallAfter(self.text.AppendText, text)
            else:
                if echo:
                    gui_log(text)
                if get_debug_console():
                    self.text.AppendText(text)

    def OnCloseWindow(self, event):  # noqa
        """Close logging console."""

        if self.frame is not None:
            self.frame.Destroy()
        self.frame = None
        self.text = None
        self.parent = None
        if get_debug_console():
            set_debug_console(False)
            log.set_echo(False)
            debug("**Debug Console Closed**\n")


class CustomApp(wx.App):
    """Custom app that adds a number of features."""

    outputWindowClass = GuiLog

    def __init__(self, *args, **kwargs):
        """
        Init the custom app.

        Provide two new inputs:
            single_instance_name: this creates an instance id with the name given
                                  this will allow you to check if this is the only
                                  instance currently open with the same name.
            callback: A callback you can do if instance checks out
        """

        self.instance = None
        self.custom_init(*args, **kwargs)
        if "single_instance_name" in kwargs:
            del kwargs["single_instance_name"]
        if "callback" in kwargs:
            del kwargs["callback"]
        wx.App.__init__(self, *args, **kwargs)

    def custom_init(self, *args, **kwargs):
        """Parse for new inputs and store them because they must be removed."""

        self.single_instance = None
        self.init_callback = None
        instance_name = kwargs.get("single_instance_name", None)
        callback = kwargs.get("callback", None)
        if instance_name is not None and isinstance(instance_name, util.string_type):
            self.single_instance = instance_name
        if callback is not None and hasattr(callback, '__call__'):
            self.init_callback = callback

    def ensure_single_instance(self, name):
        """Check to see if this is the only instance."""

        self.name = "%s-%s" % (self.single_instance, wx.GetUserId())
        self.instance = wx.SingleInstanceChecker(self.name)
        if self.instance.IsAnotherRunning():
            # wx.MessageBox("Only one instance allowed!", "ERROR", wx.OK | wx.ICON_ERROR)
            return False
        return True

    def is_instance_okay(self):
        """Return whether this is the only instance."""

        return self.instance_okay

    def OnInit(self):  # noqa
        """
        Execute callback if instance is okay.

        Store instance check variable.
        """

        self.locale = wx.Locale(wx.LANGUAGE_DEFAULT)

        self.instance_okay = True
        if self.single_instance is not None:
            if not self.ensure_single_instance(self.single_instance):
                self.instance_okay = False
        if self.init_callback is not None and self.instance_okay:
            self.init_callback()
        return True

    def OnExit(self):  # noqa
        """Cleanup instance check."""

        return 0


class ArgPipeThread(object):
    """Argument pipe thread for receiving arguments from another instance."""

    def __init__(self, app, pipe_name):
        """Init pipe thread variables."""

        self.app = app
        self.pipe_name = pipe_name

    def Start(self):  # noqa
        """Start listening to the pipe."""

        self.check_pipe = True
        self.running = True
        thread.start_new_thread(self.Run, ())

    def Stop(self):  # noqa
        """
        Stop listening to the pipe.

        Send a new line to kick the listener out from waiting.
        """

        self.check_pipe = False
        if util.platform() == "windows":
            file_handle = ctypes.windll.kernel32.CreateFileW(
                self.pipe_name,
                GENERIC_READ | GENERIC_WRITE,
                0, None,
                OPEN_EXISTING,
                0, None
            )
            data = '\n'
            bytes_written = ctypes.c_ulong(0)
            ctypes.windll.kernel32.WriteFile(
                file_handle, ctypes.c_wchar_p(data), len(data), ctypes.byref(bytes_written), None
            )
            ctypes.windll.kernel32.CloseHandle(file_handle)
        else:
            # It's okay if the pipe is broken, our goal is just to break the
            # wait loop for recieving pipe data.
            try:
                with codecs.open(self.pipe_name, "w", encoding="utf-8") as pipeout:
                    try:
                        pipeout.write('\n')
                    except IOError:
                        pass
            except util.CommonBrokenPipeError:
                pass

    def IsRunning(self):  # noqa
        """Return if the thread is still busy."""

        return self.running

    def Run(self):  # noqa
        """The actual thread listening loop."""

        if util.platform() == "windows":
            data = ""
            p = ctypes.windll.kernel32.CreateNamedPipeW(
                self.pipe_name,
                PIPE_ACCESS_DUPLEX,
                PIPE_TYPE_MESSAGE | PIPE_WAIT,
                1, 65536, 65536, 300, None
            )
            while self.check_pipe:
                ctypes.windll.kernel32.ConnectNamedPipe(p, None)
                result = ctypes.create_unicode_buffer(4096)
                bytes_read = ctypes.c_ulong(0)
                success = ctypes.windll.kernel32.ReadFile(p, result, 4096, ctypes.byref(bytes_read), None)
                if success:
                    data += result.value.replace("\r", "")
                    if len(data) and data[-1] == "\n":
                        lines = data.rstrip("\n").split("\n")
                        try:
                            args = json.loads(lines[-1])
                        except Exception:
                            args = []
                        evt = PipeEvent(data=args)
                        wx.PostEvent(self.app, evt)
                        data = ""
                ctypes.windll.kernel32.DisconnectNamedPipe(p)
                time.sleep(0.2)
            ctypes.windll.kernel32.CloseHandle(p)
        else:
            if os.path.exists(self.pipe_name):
                os.unlink(self.pipe_name)
            if not os.path.exists(self.pipe_name):
                os.mkfifo(self.pipe_name)

            while self.check_pipe:
                with codecs.open(self.pipe_name, "r", 'utf-8') as pipein:
                    while self.check_pipe:
                        line = pipein.readline()[:-1]
                        if line != "":
                            try:
                                args = json.loads(line)
                            except Exception:
                                args = []
                            evt = PipeEvent(data=args)
                            wx.PostEvent(self.app, evt)
                            break
                        time.sleep(0.2)
        self.running = False


class PipeApp(CustomApp):
    """Pip app variant that allows the app to be sent data via a pipe."""

    def __init__(self, *args, **kwargs):
        """Parse pipe args."""

        self.active_pipe = False
        self.pipe_thread = None
        self.pipe_name = kwargs.get("pipe_name", None)
        if "pipe_name" in kwargs:
            del kwargs["pipe_name"]
        CustomApp.__init__(self, *args, **kwargs)

    def OnInit(self):  # noqa
        """
        Check if this is the first instance, and if so, start the pipe listener.

        If not, send the current args to the pipe to be read
        by the first instance.
        """

        CustomApp.OnInit(self)
        self.Bind(EVT_PIPE_ARGS, self.on_pipe_args)
        if self.pipe_name is not None:
            if self.is_instance_okay():
                self.receive_arg_pipe()
            else:
                self.send_arg_pipe()
                return False
        return True

    def get_sys_args(self):
        """Get system args as unicode."""

        return util.to_unicode_argv()[1:]

    def send_arg_pipe(self):
        """Send the current arguments down the pipe."""
        argv = self.get_sys_args()
        if util.platform() == "windows":
            args = self.process_args(argv)
            file_handle = ctypes.windll.kernel32.CreateFileW(
                self.pipe_name,
                GENERIC_READ | GENERIC_WRITE,
                0, None,
                OPEN_EXISTING,
                0, None
            )
            data = json.dumps(args) + '\n'
            bytes_written = ctypes.c_ulong(0)
            ctypes.windll.kernel32.WriteFile(
                file_handle, ctypes.c_wchar_p(data), len(data) * 2, ctypes.byref(bytes_written), None
            )
            ctypes.windll.kernel32.CloseHandle(file_handle)
        else:
            with codecs.open(self.pipe_name, "w", encoding="utf-8") as pipeout:
                args = self.process_args(argv)
                pipeout.write(json.dumps(args) + '\n')

    def process_args(self, arguments):
        """Noop, but can be overriden to process the args."""

        return arguments

    def receive_arg_pipe(self):
        """Start the pipe listenr thread."""

        self.active_pipe = True
        self.pipe_thread = ArgPipeThread(self, self.pipe_name)
        self.pipe_thread.Start()

    def OnExit(self):  # noqa
        """Stop the thread if needed."""

        if self.active_pipe:
            wx.Yield()
            self.pipe_thread.Stop()
            running = True
            while running:
                running = self.pipe_thread.IsRunning()
                time.sleep(0.1)
        return CustomApp.OnExit(self)

    def on_pipe_args(self, event):
        """An overridable event for when pipe arguments are received."""

        event.Skip()


class DebugFrameExtender(object):
    """Extend frame with debugger."""

    def set_keybindings(self, keybindings=[], debug_event=None):
        """
        Method to easily set key bindings.

        Also sets up debug keybindings and events.
        """

        # Create keybinding to open debug console, bind debug console to ctrl/cmd + ` depending on platform
        # if an event is passed in.
        tbl = []
        bindings = keybindings
        if debug_event is not None:
            mod = wx.ACCEL_CMD if util.platform() == "osx" else wx.ACCEL_CTRL
            bindings.append((mod, ord('`'), debug_event))

        for binding in keybindings:
            keyid = wx.NewId()
            self.Bind(wx.EVT_MENU, binding[2], id=keyid)
            tbl.append((binding[0], binding[1], keyid))

        if len(bindings):
            self.SetAcceleratorTable(wx.AcceleratorTable(tbl))

    def open_debug_console(self):
        """Open the debug console."""

        set_debug_console(True)
        # echo out log to console
        log.set_echo(True)
        app = wx.GetApp()
        if app.stdioWin is not None:
            if app.stdioWin.frame is None:
                app.stdioWin.write(log.read(), False)
            else:
                app.stdioWin.write("", False)
            debug("**Debug Console Opened**")

    def toggle_debug_console(self):
        """Open up the debug console if closed or close if opened."""

        set_debug_console(not get_debug_console())
        app = wx.GetApp()
        if get_debug_console():
            # echo out log to console
            log.set_echo(True)
            if app.stdioWin is not None:
                app.stdioWin.write(log.read(), False)
                debug("**Debug Console Opened**")
        else:
            debug("**Debug Console Closed**")
            # disable echoing of log to console
            log.set_echo(False)
            if app.stdioWin is not None:
                app.stdioWin.close()

    def close_debug_console(self):
        """
        On close, ensure that console is closes.

        Also, make sure echo is off in case logging
        occurs after App closing.
        """

        if get_debug_console():
            debug("**Debug Console Closed**")
            set_debug_console(False)
            log.set_echo(False)
            app = wx.GetApp()
            if app.stdioWin is not None:
                app.stdioWin.close()


def _log_struct(obj, log_func, label="Object"):
    """Base logger to log a dict in pretty print format."""

    log_func(obj, log_fmt="%(loglevel)s: " + label + ": %(message)s\n", msg_fmt=json_fmt)


def json_fmt(obj):
    """Format the dict as JSON."""

    return json.dumps(obj, sort_keys=True, indent=4, separators=(',', ': '))


def gui_log(msg):
    """Logger used in the GUI frames."""

    log._log(msg, echo=False)


def debug(msg, echo=True, log_fmt="%(loglevel)s: %(message)s\n", msg_fmt=None):
    """Debug level log."""

    if get_debug_mode():
        log.debug(msg, echo=echo, log_fmt=log_fmt, msg_fmt=msg_fmt)


def info(msg, echo=True, log_fmt="%(loglevel)s: %(message)s\n", msg_fmt=None):
    """Info level log."""

    log.info(msg, echo=echo, log_fmt=log_fmt, msg_fmt=msg_fmt)


def critical(msg, echo=True, log_fmt="%(loglevel)s: %(message)s\n", msg_fmt=None):
    """Critical level log."""

    log.critical(msg, echo=echo, log_fmt=log_fmt, msg_fmt=msg_fmt)


def warning(msg, echo=True, log_fmt="%(loglevel)s: %(message)s\n", msg_fmt=None):
    """Warning level log."""

    log.warning(msg, echo=echo, log_fmt=log_fmt, msg_fmt=msg_fmt)


def error(msg, echo=True, log_fmt="%(loglevel)s: %(message)s\n", msg_fmt=None):
    """Error level log."""

    log.error(msg, echo=echo, log_fmt=log_fmt, msg_fmt=msg_fmt)


def debug_struct(obj, label="Object"):
    """Debug level dict log."""

    _log_struct(obj, debug, label)


def info_struct(obj, label="Object"):
    """Info level dict log."""

    _log_struct(obj, info, label)


def critical_struct(obj, label="Object"):
    """Critical level dict log."""

    _log_struct(obj, critical, label)


def warning_struct(obj, label="Object"):
    """Warning level dict log."""

    _log_struct(obj, warning, label)


def error_struct(obj, label="Object"):
    """Error level dict log."""

    _log_struct(obj, error, label)


def init_app_log(name, level=simplelog.ERROR):
    """Init the app log."""

    global log
    global last_level
    if level != simplelog.DEBUG:
        last_level = level
    simplelog.init_global_log(name, level=last_level)
    log = simplelog.get_global_log()


def set_debug_mode(value):
    """Set whether the app is in debug mode."""

    global DEBUG_MODE
    global last_level
    DEBUG_MODE = bool(value)
    current_level = log.get_level()
    if DEBUG_MODE:
        if current_level > simplelog.DEBUG:
            last_level = current_level
        log.set_level(simplelog.DEBUG)
    elif not DEBUG_MODE:
        if last_level == simplelog.DEBUG:
            last_level = simplelog.ERROR
        log.set_level(last_level)


def set_debug_console(value):
    """Set debug console enable."""

    global DEBUG_CONSOLE
    DEBUG_CONSOLE = bool(value)


def get_debug_mode():
    """Get the debug mode."""

    return DEBUG_MODE


def get_debug_console():
    """Get debug console mode."""

    return DEBUG_CONSOLE



"""Global editor dialog."""
from __future__ import unicode_literals
import wx
from . import basic_dialogs
from . import gui
from . import settings_key_bindings
from . import settings_codes as sc
from ..rgba import RGBA
from .. import util


class GlobalEditor(gui.GlobalSetting, settings_key_bindings.SettingsKeyBindings):
    """GlobalEditor."""

    def __init__(self, parent, current_entries, name, value, insert=False):
        """Initialize."""

        super(GlobalEditor, self).__init__(parent)

        if util.platform() == "windows":
            self.SetDoubleBuffered(True)

        self.setup_keybindings()
        self.Fit()
        size = self.GetSize()
        self.SetMinSize(size)
        size.Set(-1, size[1])
        self.SetMaxSize(size)
        self.obj_key = name
        self.obj_val = value
        self.color_save = ""
        self.apply_settings = False
        self.color_setting = False
        self.m_color_picker.Disable()
        self.entries = current_entries
        self.current_name = name
        self.valid = True
        self.insert = bool(insert)

        self.m_name_textbox.SetValue(self.obj_key)
        try:
            RGBA(self.obj_val)
            self.color_setting = True
            self.color_save = self.obj_val
            self.m_color_picker.Enable()
            self.m_color_checkbox.SetValue(True)
        except:
            pass
        self.m_value_textbox.SetValue(self.obj_val)

    def on_color_button_click(self, event):
        """Handle color button click event."""

        if not self.color_setting:
            event.Skip()
            return
        color = None
        data = wx.ColourData()
        data.SetChooseFull(True)

        alpha = None

        text = self.m_value_textbox.GetValue()
        rgb = RGBA(text)
        if len(text) == 9:
            alpha == text[7:9]

        # set the default color in the chooser
        data.SetColour(wx.Colour(rgb.r, rgb.g, rgb.b))

        # construct the chooser
        dlg = wx.ColourDialog(self, data)

        if dlg.ShowModal() == wx.ID_OK:
            # set the panel background color
            color = dlg.GetColourData().GetColour().GetAsString(wx.C2S_HTML_SYNTAX)
            self.m_value_textbox.SetValue(color if alpha is None else color + alpha)
        dlg.Destroy()
        event.Skip()

    def on_global_checkbox(self, event):
        """Handle global checkbox event."""

        if event.IsChecked():
            self.m_color_picker.Enable()
            self.color_setting = True
            try:
                RGBA(self.m_value_textbox.GetValue())
                self.on_color_change(event)
            except:
                self.m_value_textbox.SetValue("#000000")
            return
        else:
            self.color_setting = False
            self.m_color_picker.Disable()
            self.m_color_picker.SetBackgroundColour(wx.Colour(255, 255, 255))
            self.m_color_picker.Refresh()
        event.Skip()

    def is_name_valid(self):
        """Check if name is valid."""

        valid = True
        name = self.m_name_textbox.GetValue()
        if name != self.current_name:
            for k in self.entries:
                if name == k:
                    valid = False
                    break
        return valid

    def on_global_name_blur(self, event):
        """Handle global name blur event."""

        if not self.is_name_valid():
            basic_dialogs.errormsg(
                "Key name \"%s\" already exists in global settings. "
                "Please use a different name." % self.m_name_textbox.GetValue()
            )
            self.m_name_textbox.SetValue(self.current_name)
        else:
            self.current_name = self.m_name_textbox.GetValue()

    def on_color_change(self, event):
        """Handle color change event."""

        if not self.color_setting:
            event.Skip()
            return
        text = self.m_value_textbox.GetValue()
        try:
            cl = RGBA(text)
        except:
            event.Skip()
            return

        cl.apply_alpha(self.Parent.m_style_settings.bg_color.get_rgb())
        bg = wx.Colour(cl.r, cl.g, cl.b)
        self.m_color_picker.SetBackgroundColour(bg)
        self.m_color_picker.Refresh()

    def on_color_focus(self, event):
        """Handle color focus event."""

        if not self.color_setting:
            event.Skip()
            return
        if self.color_setting:
            self.color_save = self.m_value_textbox.GetValue()
        event.Skip()

    def on_color_blur(self, event):
        """Handle color blur event."""

        if not self.color_setting:
            event.Skip()
            return
        if self.color_setting:
            text = self.m_value_textbox.GetValue()
            try:
                RGBA(text)
            except:
                self.m_value_textbox.SetValue(self.color_save)
        event.Skip()

    def on_apply_button_click(self, event):
        """Handle apply button click event."""

        self.m_apply_button.SetFocus()
        if self.is_name_valid():
            self.apply_settings = True
            self.Close()
        else:
            basic_dialogs.errormsg(
                "Key name \"%s\" already exists in global settings. "
                "Please use a different name." % self.m_name_textbox.GetValue()
            )
            self.m_name_textbox.SetValue(self.current_name)

    def on_set_color_close(self, event):
        """Handle set color close event."""

        self.obj_key = self.m_name_textbox.GetValue()
        self.current_name = self.obj_key

        if self.apply_settings:
            self.obj_val = self.m_value_textbox.GetValue()

            if self.insert:
                grid = self.Parent.m_global_settings.m_plist_grid
                num = grid.GetNumberRows()
                row = grid.GetGridCursorRow()
                if num > 0:
                    grid.InsertRows(row, 1, True)
                else:
                    grid.AppendRows(1)
                    row = 0
                grid.GetParent().update_row(row, self.obj_key, self.obj_val)
                grid.GetParent().go_cell(grid, row, 0)
                self.Parent.update_plist(sc.ADD, {"table": "global", "index": self.obj_key, "data": self.obj_val})
            else:
                self.Parent.set_global_object(self.obj_key, self.obj_val)

        event.Skip()



"""Helper for gird objects."""
from __future__ import unicode_literals
import sys
import wx
from ..rgba import RGBA


class GridHelper(object):
    """Grid helper."""

    cell_select_semaphore = False
    range_semaphore = False
    current_row = None
    current_col = None

    def setup_keybindings(self):
        """Setup grid keybindings."""

        deleteid = wx.NewId()
        insertid = wx.NewId()

        self.Bind(wx.EVT_MENU, self.on_delete_row, id=deleteid)
        self.Bind(wx.EVT_MENU, self.on_insert_row, id=insertid)

        accel_tbl = wx.AcceleratorTable(
            [
                (wx.ACCEL_NORMAL, wx.WXK_DELETE, deleteid),
                (wx.ACCEL_CMD, ord('I'), insertid) if sys.platform == "darwin" else (wx.ACCEL_CTRL, ord('I'), insertid)
            ]
        )
        self.SetAcceleratorTable(accel_tbl)

    def go_cell(self, grid, row, col, focus=False):
        """Go to cell."""

        if focus:
            grid.GoToCell(row, col)
        else:
            grid.SetGridCursor(row, col)
        bg = grid.GetCellBackgroundColour(row, 0)
        lum = RGBA(bg.GetAsString(wx.C2S_HTML_SYNTAX)).get_luminance()
        if lum > 128:
            bg.Set(0, 0, 0)
        else:
            bg.Set(255, 255, 255)
        grid.SetCellHighlightColour(bg)

    def mouse_motion(self, event):
        """Capture if mouse is dragging."""

        if event.Dragging():       # mouse being dragged?
            pass                   # eat the event
        else:
            event.Skip()           # no dragging, pass on to the window

    def grid_key_down(self, event):
        """Check for certain key down events in the grid."""

        no_modifiers = event.GetModifiers() == 0
        alt_mod = event.GetModifiers() == wx.MOD_ALT
        if no_modifiers and event.GetKeyCode() == ord('B'):
            self.toggle_bold()
            return
        elif no_modifiers and event.GetKeyCode() == ord('I'):
            self.toggle_italic()
            return
        elif no_modifiers and event.GetKeyCode() == ord('U'):
            self.toggle_underline()
            return
        elif no_modifiers and event.GetKeyCode() == wx.WXK_RETURN:
            self.edit_cell()
            return
        elif alt_mod and event.GetKeyCode() == wx.WXK_UP:
            self.row_up()
            return
        elif alt_mod and event.GetKeyCode() == wx.WXK_DOWN:
            self.row_down()
            return
        elif alt_mod and event.GetKeyCode() == wx.WXK_LEFT:
            self.on_panel_left(event)
            return
        elif alt_mod and event.GetKeyCode() == wx.WXK_RIGHT:
            self.on_panel_right(event)
            return
        elif event.AltDown():
            # Eat...NOM NOM
            if event.GetKeyCode() == wx.WXK_UP:
                return
            elif event.GetKeyCode() == wx.WXK_DOWN:
                return
            elif event.GetKeyCode() == wx.WXK_LEFT:
                return
            elif event.GetKeyCode() == wx.WXK_RIGHT:
                return
        elif event.ShiftDown():
            # Eat...NOM NOM
            if event.GetKeyCode() == wx.WXK_UP:
                return
            elif event.GetKeyCode() == wx.WXK_DOWN:
                return
            elif event.GetKeyCode() == wx.WXK_LEFT:
                return
            elif event.GetKeyCode() == wx.WXK_RIGHT:
                return
        event.Skip()

    def grid_select_cell(self, event):
        """Grid cell selected."""

        grid = self.m_plist_grid
        if not self.cell_select_semaphore and event.Selecting():
            self.cell_select_semaphore = True
            self.current_row = event.GetRow()
            self.current_col = event.GetCol()
            self.go_cell(grid, self.current_row, self.current_col)
            self.cell_select_semaphore = False
        event.Skip()

    def on_panel_left(self, event):
        """Handle left key press."""

        grid = self.m_plist_grid
        grid.GetParent().GetParent().ChangeSelection(0)
        grid.GetParent().GetParent().GetPage(0).m_plist_grid.SetFocus()

    def on_panel_right(self, event):
        """Handle right key press."""

        grid = self.m_plist_grid
        grid.GetParent().GetParent().ChangeSelection(1)
        grid.GetParent().GetParent().GetPage(1).m_plist_grid.SetFocus()

    def on_row_up(self, event):
        """Handle row up."""

        self.row_up()

    def on_row_down(self, event):
        """Handle row down."""

        self.row_down()

    def on_insert_row(self, event):
        """Handle insert row."""

        self.insert_row()

    def on_delete_row(self, event):
        """Handle delete row."""

        self.delete_row()

    def on_edit_cell_key(self, event):
        """Handle edit cell key."""

        self.edit_cell()

    def on_toggle_bold(self, event):
        """Handle bold event."""

        self.toggle_bold()

    def on_toggle_italic(self, event):
        """Handle italic event."""

        self.toggle_italic()

    def on_toggle_underline(self, event):
        """Handle underline event."""

        self.toggle_underline()

    def toggle_bold(self):
        """Override for toggle bold."""

    def toggle_italic(self):
        """Override for toggle italic."""

    def toggle_underline(self):
        """Override for toggle underline."""

    def row_up(self):
        """Override row up."""

    def row_down(self):
        """Override for row down."""

    def edit_cell(self):
        """Override for edit cell."""

    def delete_row(self):
        """Override for delete row."""

    def insert_row(self):
        """Override for insert tow."""
